from collections import defaultdict

from whistle.errors import RemovedInWhistle2Error
from whistle.typing import IListener


class ListenersCollection:
    def __init__(self) -> None:
        self._items = defaultdict(lambda: defaultdict(list))
        self._sorted = {}

    def add(self, event_id: str, listener: IListener, /, *, priority: int = 0) -> None:
        """
        Add a listener for the given event id, with the given priority.

        Priority must be a number between -127 and 128, inclusive. As a convention coming from unix process priorities,
        we use priorities ranging from -20 (highest priority) to 20 (lowest priority), with 0 being the default
        priority.

        The actual permited range is so because python dicts are WAY faster with string keys than integers. This fact
        still puzzles me but the difference is huge. The same rationale applies to the event_id parameter: although we
        could support arbitrary hashable values as event ids (and we encouraged so in whistle 1.x), we now restrict it
        to strings for performance reasons.

        :param event_id: string identifier for the event
        :param listener: callback to be called when the event is dispatched
        :param priority: integer priority for the listener (-20 (high) to 19 (low))

        """

        if not callable(listener):
            raise TypeError(f"Listener should be a callable, {type(listener)} given")

        # priority = chr(priority)
        self._items[event_id][priority].append(listener)
        if event_id in self._sorted:
            del self._sorted[event_id]

    def get(self, event_id: str, /) -> tuple[IListener, ...]:
        """
        Gets the listeners for the given event id, in order of priority.

        :param event_id: string identifier for the event

        """
        if event_id is None:
            raise RemovedInWhistle2Error("ListenersCollection.get() without event_id is not accepted anymore.")

        if event_id not in self._items:
            return ()

        if event_id not in self._sorted:
            self._sort(event_id)

        return self._sorted[event_id]

    def has(self, event_id: str, /) -> bool:
        # todo avoid sort for has !!!
        return bool(len(self.get(event_id)))

    def remove(self, event_id: str, listener: IListener, /) -> None:
        """ """
        if event_id not in self._items:
            raise ValueError(f"Listener {listener} is not registered for event {event_id}.")

        for priority, listeners in self._items[event_id].items():
            if listener in listeners:
                self._items[event_id][priority] = list(filter(lambda _listener: _listener != listener, listeners))

        if event_id in self._sorted:
            del self._sorted[event_id]

    def keys(self):
        return self._items.keys()

    def items(self):
        for event_id in self.keys():
            yield event_id, self.get(event_id)

    def _sort(self, event_id, /):
        # set to an empty list just in case some concurrent access happens
        _sorted = ()
        self._sorted[event_id] = _sorted

        for priority in sorted(self._items.get(event_id, {}).keys()):
            _sorted += tuple(self._items[event_id][priority])

        self._sorted[event_id] = _sorted



# Configuration file for the Sphinx documentation builder.

import builtins
import os
import sys
from datetime import datetime

builtins.__sphinx__ = True

sys.path.insert(0, os.path.abspath(".."))

project = "Whistle"

first_year, current_year = 2015, datetime.now().year
author = "Romain Dorgueil"
copyright = f"{current_year}, {author}"
if current_year > first_year:
    copyright = str(first_year) + "-" + copyright
version = release = ".".join(__import__("whistle").__version__.split(".")[0:2])


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.graphviz",
    "sphinx.ext.ifconfig",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_sitemap",
    "sphinxcontrib.jquery",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

root_doc = "index"

html_theme = "furo"
html_static_path = ["_static"]
html_theme_options = {
    "light_logo": "whistle.png",
}
html_js_files = ["js/links-target-blank.js"]
html_css_files = ["css/whistle.css"]
html_baseurl = "https://python-whistle.readthedocs.io/latest/"


todo_include_todos = True
html_show_sphinx = False

autodoc_typehints = "description"
autodoc_member_order = "groupwise"
autodoc_default_flags = ["members", "undoc-members", "show-inheritance"]
autodoc_class_signature = "separated"

autoclass_content = "both"
add_module_names = False
pygments_style = "sphinx"
graphviz_output_format = "svg"

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}



"""
Listener management and introspection.

This example demonstrates how to check for listeners, retrieve them,
and remove them when needed.
"""

from whistle import EventDispatcher


def main():
    """Demonstrate listener management operations."""
    dispatcher = EventDispatcher()

    def listener_one(event):
        print("Listener one executed")

    def listener_two(event):
        print("Listener two executed")

    # Check if event has listeners before registering
    print("=== Initial State ===")
    print(f"Has listeners for 'app.start': {dispatcher.has_listeners('app.start')}")

    # Register listeners
    dispatcher.add_listener("app.start", listener_one)
    dispatcher.add_listener("app.start", listener_two)

    print("\n=== After Registration ===")
    print(f"Has listeners for 'app.start': {dispatcher.has_listeners('app.start')}")

    # Get all listeners for an event
    listeners = dispatcher.get_listeners("app.start")
    print(f"Number of listeners: {len(listeners)}")

    # Dispatch event
    print("\n=== Dispatching Event ===")
    dispatcher.dispatch("app.start")

    # Remove a listener
    print("\n=== After Removing listener_two ===")
    dispatcher.remove_listener("app.start", listener_two)
    print(f"Number of listeners: {len(dispatcher.get_listeners('app.start'))}")

    # Dispatch again - only listener_one executes
    print("\n=== Dispatching Event Again ===")
    dispatcher.dispatch("app.start")

    # Check if any events have listeners (v1.x compatibility)
    print("\n=== Global Check ===")
    print(f"Dispatcher has any listeners: {dispatcher.has_listeners()}")


if __name__ == "__main__":
    main()



import asyncio
from typing import Optional, cast

from whistle.dispatchers.base import AbstractEventDispatcher
from whistle.event import Event
from whistle.typing import IDispatchedEvent, IEvent, IListener


class AsyncEventDispatcher(AbstractEventDispatcher):
    """
    Adapts whiste's EventDispatcher to be async.

    .. versionadded:: 2.0

    """

    def add_listener(self, event_id: str, listener: IListener, /, *, priority: int = 0):
        # Use asyncio.iscoroutinefunction as it's more lenient and handles mock objects better
        if not asyncio.iscoroutinefunction(listener):
            raise TypeError(f"Listener should be a coroutine function, {type(listener)} given")
        return super().add_listener(event_id, listener, priority=priority)

    def dispatch(self, event_id: str, event: Optional[IEvent] = None, /) -> IDispatchedEvent:
        raise NotImplementedError("AsyncEventDispatcher does not implement sync dispatch")

    async def adispatch(self, event_id: str, event: Optional[IEvent] = None, /) -> IDispatchedEvent:
        """
        :param event_id:
        :param event:
        :return:
        """
        if event is None:
            event = Event()

        event.name = event_id
        event.dispatcher = self

        await self._adispatch(self.get_listeners(event_id), event)
        return cast(IDispatchedEvent, event)

    async def _adispatch(self, listeners, event):
        for listener in listeners:
            await listener(event)
            if event.propagation_stopped:
                break



"""
Basic asynchronous event dispatching.

This example demonstrates async/await usage with AsyncEventDispatcher.
All listeners must be async functions (coroutines).
"""

import asyncio

from whistle import AsyncEventDispatcher


async def main():
    """Demonstrate basic asynchronous event dispatching."""
    # Create an async event dispatcher
    dispatcher = AsyncEventDispatcher()

    # Define an async listener function
    async def on_data_received(event):
        # Simulate async operation
        await asyncio.sleep(0.01)
        print(f"Data received for event: {event.name}")

    # Register the async listener
    dispatcher.add_listener("data.received", on_data_received)

    # Dispatch the event asynchronously
    await dispatcher.adispatch("data.received")


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
from typing import Optional, cast

from whistle.dispatchers.base import AbstractEventDispatcher
from whistle.event import Event
from whistle.typing import IDispatchedEvent, IEvent, IListener


class EventDispatcher(AbstractEventDispatcher):
    """
    Main class of the library, it is responsible for keeping track of registered listeners, and dispatching events to
    them. All listeners are scoped to the event dispatcher instance, so you can have multiple event dispatchers with
    different sets of listeners.

    """

    def add_listener(self, event_id: str, listener: IListener, /, *, priority: int = 0):
        # Use asyncio.iscoroutinefunction as it's more lenient and handles mock objects better
        if asyncio.iscoroutinefunction(listener):
            raise TypeError(f"Listener should not be a coroutine function, {type(listener)} given")
        return super().add_listener(event_id, listener, priority=priority)

    def dispatch(self, event_id: str, event: Optional[IEvent] = None, /) -> IDispatchedEvent:
        """
        Dispatch the given event, with the given event id.

        An optional event can be given, and should respect the :class:`whistle.protocols.IEvent` protocol.
        If no event is given, a new :class:`whistle.event.Event` instance is created and used.

        Returns the event instance after it has been dispatched, whether it has been created or provided by the caller.

        :param event_id: hashable identifier for the event
        :param event: optional event instance
        :return: the event instance after it has been dispatched

        """
        if event is None:
            event = Event()

        # todo should name be part of dispatched event ?
        event.name = event_id
        event.dispatcher = self

        self.do_dispatch(self._listeners.get(event_id), event)

        return cast(IDispatchedEvent, event)

    async def adispatch(self, event_id: str, event: Optional[IEvent] = None, /) -> IDispatchedEvent:
        # allows to use the async interface with a sync dispatcher, although this is not recommended
        # todo add a strict mode that raises an error when trying to use async interface with a sync dispatcher
        return self.dispatch(event_id, event)

    def do_dispatch(self, listeners, event: IEvent, /):
        for listener in listeners:
            listener(event)
            if event.propagation_stopped:
                break



from .dispatcher import IAsyncEventDispatcher, IDispatchedEvent, IEventDispatcher
from .event import IEvent
from .listener import IListener

__all__ = [
    "IAsyncEventDispatcher",
    "IDispatchedEvent",
    "IEvent",
    "IEventDispatcher",
    "IListener",
]



from whistle._version import __version__
from whistle.dispatchers import AsyncEventDispatcher, EventDispatcher
from whistle.event import Event
from whistle.typing import *
from whistle.typing import __all__ as _typing_all

__all__ = [
    "AsyncEventDispatcher",
    "Event",
    "EventDispatcher",
    "__version__",
    *_typing_all,
]



import random
from random import randint
from unittest.mock import Mock

import pytest

from whistle.listeners import ListenersCollection


def test_interface():
    assert set(filter(lambda x: not x.startswith("_"), dir(ListenersCollection))) == {
        "add",
        "get",
        "has",
        "remove",
        "keys",
        "items",
    }


@pytest.mark.parametrize("event_id", ["event", 42, object()])
def test_sort_listeners(event_id):
    coll = ListenersCollection()

    a, b, c, d, unrelated = Mock(), Mock(), Mock(), Mock(), Mock()

    coll.add(event_id, a)
    coll.add(event_id, b)
    coll.add(event_id, c)

    coll.add("unrelated", unrelated)

    assert coll.get(event_id) == (a, b, c)
    assert coll.has(event_id)
    assert coll.has("unrelated")
    assert not coll.has("something else")
    assert coll.keys() == {event_id, "unrelated"}

    coll.add(event_id, d, priority=-1)
    assert coll.get(event_id) == (d, a, b, c)
    assert coll.has(event_id)
    assert coll.has("unrelated")


@pytest.mark.benchmark
def test_add_benchmark(benchmark):
    coll = ListenersCollection()

    @benchmark
    def add():
        coll.add(f"event{randint(1, 10)}", lambda e: None)


@pytest.mark.benchmark
def test_get_benchmark(benchmark):
    coll = ListenersCollection()
    event_ids = set()
    for i in range(1000):
        event_id = "event_randint(1,10)"
        coll.add(event_id, lambda e: None)
        event_ids.add(event_id)
    event_ids = list(event_ids)

    @benchmark
    def get():
        coll.get(random.choice(event_ids))


@pytest.mark.benchmark
def test_e2e_benchmark(benchmark):
    listeners = {f"event_{i}": Mock() for i in range(1000)}
    coll = ListenersCollection()
    i = 0
    event_ids = list(listeners.keys())

    @benchmark
    def add_get():
        nonlocal i
        event_id = random.choice(event_ids)
        coll.add(event_id, lambda e: None)
        if i % 10 == 0:
            coll.get(event_id)
        i += 1



"""
Using the @listen decorator for cleaner listener registration.

This example demonstrates the decorator pattern for registering listeners,
which provides a more elegant syntax.
"""

from whistle import EventDispatcher


def main():
    """Demonstrate listener registration using decorators."""
    dispatcher = EventDispatcher()

    # Use decorator to register listeners
    @dispatcher.listen("user.login")
    def on_user_login(event):
        print(f"User logged in - event: {event.name}")

    @dispatcher.listen("user.logout")
    def on_user_logout(event):
        print(f"User logged out - event: {event.name}")

    # Dispatch events
    dispatcher.dispatch("user.login")
    dispatcher.dispatch("user.logout")


if __name__ == "__main__":
    main()



from abc import ABCMeta, abstractmethod
from typing import Optional

from whistle.listeners import ListenersCollection
from whistle.typing import IDispatchedEvent, IEvent, IListener


class AbstractEventDispatcher(metaclass=ABCMeta):
    def __init__(self):
        self._listeners = ListenersCollection()

    def get_listeners(self, event_id: Optional[str] = None, /):
        # compatibility with 1.x
        if event_id is None:
            return {event_id: self._listeners.get(event_id) for event_id in self._listeners.keys()}

        return self._listeners.get(event_id)

    def has_listeners(self, event_id: Optional[str] = None, /):
        # compatibility with 1.x
        if event_id is None:
            return bool(len(self._listeners.keys()))

        return self._listeners.has(event_id)

    def add_listener(self, event_id: str, listener: IListener, /, *, priority: int = 0):
        return self._listeners.add(event_id, listener, priority=priority)

    def remove_listener(self, event_id: str, listener: IListener, /):
        # deprecated compatibility method
        return self._listeners.remove(event_id, listener)

    def listen(self, event_id: str, /, *, priority=0):
        """
        Decorator that add the decorated functions as one of this instance listeners, for the given event name.

        deperecated ?

        :param event_id:
        :param priority:
        :return:

        """

        def wrapper(listener):
            self._listeners.add(event_id, listener, priority=priority)
            return listener

        return wrapper

    @abstractmethod
    def dispatch(self, event_id: str, event: Optional[IEvent] = None, /) -> IDispatchedEvent: ...

    @abstractmethod
    async def adispatch(self, event_id: str, event: Optional[IEvent] = None, /) -> IDispatchedEvent: ...



"""
Basic synchronous event dispatching example.

This example demonstrates the simplest use case: creating a dispatcher,
registering a listener, and dispatching an event.
"""

from whistle import EventDispatcher


def main():
    """Demonstrate basic synchronous event dispatching."""
    # Create an event dispatcher
    dispatcher = EventDispatcher()

    # Define a listener function
    def on_spectacle_starts(event):
        print("Please turn down your phones!")

    # Register the listener for a specific event
    dispatcher.add_listener("spectacle.starts", on_spectacle_starts)

    # Dispatch the event
    dispatcher.dispatch("spectacle.starts")


if __name__ == "__main__":
    main()



"""
Test suite for documentation examples.

These tests ensure that all example code in the documentation works correctly.
Following TDD principles, these tests are written first, then examples are implemented to pass them.
"""

from io import StringIO
from unittest import mock


class TestBasicExamples:
    """Test basic synchronous examples."""

    def test_01_basic_sync(self):
        """Test basic synchronous dispatcher example."""
        from examples import example_01_basic_sync

        # Should run without errors
        with mock.patch("sys.stdout", new_callable=StringIO) as mock_stdout:
            example_01_basic_sync.main()
            output = mock_stdout.getvalue()

        # Should print something (event was dispatched)
        assert len(output) > 0

    def test_02_sync_multiple_listeners(self):
        """Test synchronous example with multiple listeners."""
        from examples import example_02_sync_multiple_listeners

        with mock.patch("sys.stdout", new_callable=StringIO) as mock_stdout:
            example_02_sync_multiple_listeners.main()
            output = mock_stdout.getvalue()

        # Should show multiple listeners executed
        assert len(output.splitlines()) >= 2

    def test_03_sync_decorator(self):
        """Test synchronous example using @listen decorator."""
        from examples import example_03_sync_decorator

        with mock.patch("sys.stdout", new_callable=StringIO) as mock_stdout:
            example_03_sync_decorator.main()
            output = mock_stdout.getvalue()

        # Should show decorator-registered listener executed
        assert len(output) > 0


class TestAsyncExamples:
    """Test asynchronous examples."""

    async def test_04_basic_async(self):
        """Test basic asynchronous dispatcher example."""
        from examples import example_04_basic_async

        with mock.patch("sys.stdout", new_callable=StringIO) as mock_stdout:
            await example_04_basic_async.main()
            output = mock_stdout.getvalue()

        # Should print something (async event was dispatched)
        assert len(output) > 0

    async def test_05_async_io_operations(self):
        """Test async example with I/O operations."""
        from examples import example_05_async_io_operations

        with mock.patch("sys.stdout", new_callable=StringIO) as mock_stdout:
            await example_05_async_io_operations.main()
            output = mock_stdout.getvalue()

        # Should show async operations completed
        assert len(output) > 0


class TestAdvancedExamples:
    """Test advanced feature examples."""

    def test_06_priorities(self):
        """Test priority system example."""
        from examples import example_06_priorities

        with mock.patch("sys.stdout", new_callable=StringIO) as mock_stdout:
            example_06_priorities.main()
            output = mock_stdout.getvalue()

        lines = output.splitlines()
        # Should show listeners executed in priority order
        assert len(lines) >= 3

        # Verify priority order (high priority listeners run first)
        # High priority should appear before low priority in output
        output_str = " ".join(lines)
        high_pos = output_str.find("high")
        low_pos = output_str.find("low")
        if high_pos >= 0 and low_pos >= 0:
            assert high_pos < low_pos, "High priority should execute before low priority"

    def test_07_propagation(self):
        """Test event propagation control example."""
        from examples import example_07_propagation

        with mock.patch("sys.stdout", new_callable=StringIO) as mock_stdout:
            example_07_propagation.main()
            output = mock_stdout.getvalue()

        # Should show some listeners executed, but not all (propagation stopped)
        assert len(output) > 0

    def test_08_custom_event(self):
        """Test custom event class example."""
        from examples import example_08_custom_event

        with mock.patch("sys.stdout", new_callable=StringIO) as mock_stdout:
            example_08_custom_event.main()
            output = mock_stdout.getvalue()

        # Should show custom event data being used
        assert len(output) > 0

    def test_09_listener_management(self):
        """Test listener management operations example."""
        from examples import example_09_listener_management

        with mock.patch("sys.stdout", new_callable=StringIO) as mock_stdout:
            example_09_listener_management.main()
            output = mock_stdout.getvalue()

        # Should demonstrate has_listeners, get_listeners, remove_listener
        assert len(output) > 0


class TestPatternExamples:
    """Test common pattern examples."""

    def test_10_patterns(self):
        """Test common patterns example."""
        from examples import example_10_patterns

        with mock.patch("sys.stdout", new_callable=StringIO) as mock_stdout:
            example_10_patterns.main()
            output = mock_stdout.getvalue()

        # Should demonstrate multiple patterns
        assert len(output) > 0
        # Should show different pattern sections
        assert len(output.splitlines()) >= 4



class Event(object):
    """
    Base class to represent whistle's events. You can subclass this if you want to embed special data and associated
    logic with your events, or just let the event dispatcher create instances for you

    The event handlers will have :class:`Event` instances passed, so you can bundle any data required by your handlers
    there.

    """

    name = None
    """Event name placeholder, will be set by dispatcher."""

    dispatcher = None

    propagation_stopped = False
    """Has the event propagation ended?"""

    def stop_propagation(self):
        """Stop event propagation, meaning that the remaining handlers won't be called after this one."""
        self.propagation_stopped = True



from unittest.mock import AsyncMock, Mock

import pytest
from pytest import fixture

from whistle import AsyncEventDispatcher, EventDispatcher


class BaseDispatcherTest:
    @fixture
    def dispatcher(self):
        raise NotImplementedError()

    def create_handler(self):
        raise NotImplementedError()

    def create_invalid_handler(self):
        raise NotImplementedError()

    def test_dispatch_invalid(self, dispatcher):
        handler = self.create_invalid_handler()
        with pytest.raises(TypeError):
            dispatcher.add_listener("test", handler)


class TestAsyncDispatcher(BaseDispatcherTest):
    @fixture
    def dispatcher(self):
        return AsyncEventDispatcher()

    def create_handler(self):
        return AsyncMock()

    def create_invalid_handler(self):
        return Mock()

    def test_dispatch(self, dispatcher):
        handler = self.create_handler()
        dispatcher.add_listener("test", handler)
        with pytest.raises(NotImplementedError):
            dispatcher.dispatch("test")

    async def test_adispatch(self, dispatcher):
        handler = self.create_handler()
        dispatcher.add_listener("test", handler)
        await dispatcher.adispatch("test")
        assert handler.called


class TestSyncDispatcher(BaseDispatcherTest):
    @fixture
    def dispatcher(self):
        return EventDispatcher()

    def create_handler(self):
        return Mock()

    def create_invalid_handler(self):
        return AsyncMock()

    def test_dispatch(self, dispatcher):
        handler = self.create_handler()
        dispatcher.add_listener("test", handler)
        dispatcher.dispatch("test")
        assert handler.called

    async def test_adispatch(self, dispatcher):
        handler = self.create_handler()
        dispatcher.add_listener("test", handler)
        await dispatcher.adispatch("test")
        assert handler.called



"""
Common patterns and recipes.

This example demonstrates several common patterns for using event dispatchers
in real-world applications.
"""

from whistle import Event, EventDispatcher


def main():
    """Demonstrate common event dispatcher patterns."""

    # Pattern 1: Conditional Dispatch (check before dispatching)
    print("=== Pattern 1: Conditional Dispatch ===")
    dispatcher = EventDispatcher()

    @dispatcher.listen("optional.event")
    def optional_handler(event):
        print("Optional handler executed")

    # Only dispatch if listeners are registered
    if dispatcher.has_listeners("optional.event"):
        dispatcher.dispatch("optional.event")
        print("Event dispatched")
    else:
        print("No listeners, skipping dispatch")

    # Pattern 2: Plugin Architecture
    print("\n=== Pattern 2: Plugin Architecture ===")

    class Plugin:
        """Base plugin class."""

        def register(self, dispatcher):
            """Register plugin's event listeners."""
            raise NotImplementedError

    class LoggingPlugin(Plugin):
        """Plugin that logs events."""

        def register(self, dispatcher):
            dispatcher.add_listener("*", self.log_event, priority=-20)

        def log_event(self, event):
            print(f"[LOG] Event: {event.name}")

    class MetricsPlugin(Plugin):
        """Plugin that tracks metrics."""

        def register(self, dispatcher):
            dispatcher.add_listener("*", self.track_metric, priority=-10)

        def track_metric(self, event):
            print(f"[METRICS] Tracking: {event.name}")

    # Set up plugins
    app_dispatcher = EventDispatcher()
    plugins = [LoggingPlugin(), MetricsPlugin()]

    for plugin in plugins:
        plugin.register(app_dispatcher)

    # All plugins respond to all events
    @app_dispatcher.listen("*")
    def business_logic(event):
        print(f"[BUSINESS] Handling: {event.name}")

    app_dispatcher.dispatch("*")

    # Pattern 3: Event-Driven Workflow
    print("\n=== Pattern 3: Event-Driven Workflow ===")
    workflow_dispatcher = EventDispatcher()

    @workflow_dispatcher.listen("order.placed", priority=-10)
    def validate_order(event):
        print("Step 1: Validating order")
        event.validated = True

    @workflow_dispatcher.listen("order.placed", priority=0)
    def charge_payment(event):
        if getattr(event, "validated", False):
            print("Step 2: Charging payment")
            event.paid = True

    @workflow_dispatcher.listen("order.placed", priority=10)
    def ship_order(event):
        if getattr(event, "paid", False):
            print("Step 3: Shipping order")

    workflow_event = Event()
    workflow_dispatcher.dispatch("order.placed", workflow_event)

    # Pattern 4: Audit Logging
    print("\n=== Pattern 4: Audit Logging ===")
    audit_dispatcher = EventDispatcher()

    class AuditEvent(Event):
        """Event with audit information."""

        def __init__(self, user_id, action):
            self.user_id = user_id
            self.action = action

    @audit_dispatcher.listen("audit.log", priority=-20)
    def write_audit_log(event):
        print(f"[AUDIT] User {event.user_id} performed: {event.action}")

    @audit_dispatcher.listen("audit.log")
    def send_compliance_report(event):
        if event.action in ["delete", "modify"]:
            print(f"[COMPLIANCE] Flagged action: {event.action}")

    audit_event = AuditEvent(user_id=42, action="delete")
    audit_dispatcher.dispatch("audit.log", audit_event)


if __name__ == "__main__":
    main()



from typing import Optional, Protocol

from .event import IEvent
from .listener import IListener


class IAbstractEventDispatcher(Protocol):
    def get_listeners(self, event_id: Optional[str] = None, /): ...

    def has_listeners(self, event_id: Optional[str] = None, /): ...

    def add_listener(self, event_id: str, listener: IListener, /, *, priority: int = 0): ...

    def remove_listener(self, event_id: str, listener: IListener, /): ...


class IDispatchedEvent(IEvent, Protocol):
    # todo: add typed var for event dispatcher type
    dispatcher: IAbstractEventDispatcher


class IEventDispatcher(IAbstractEventDispatcher, Protocol):
    def dispatch(self, event_id, event=None, /) -> IDispatchedEvent: ...


class IAsyncEventDispatcher(IAbstractEventDispatcher, Protocol):
    async def adispatch(self, event_id: str, event: Optional[IEvent] = None, /) -> IDispatchedEvent: ...



"""
Event propagation control.

This example demonstrates how to stop event propagation, preventing
remaining listeners from executing. Useful for validation or error handling.
"""

from whistle import EventDispatcher


def main():
    """Demonstrate stopping event propagation."""
    dispatcher = EventDispatcher()

    # Validator that may stop propagation
    def validate_data(event):
        print("Validating data...")
        # Simulate validation failure
        if not hasattr(event, "valid") or not event.valid:
            print("Validation failed! Stopping propagation.")
            event.stop_propagation()
        else:
            print("Validation passed")

    # This will only run if validation passes
    def process_data(event):
        print("Processing data...")

    # This will also only run if validation passes
    def save_data(event):
        print("Saving data...")

    # Register listeners in order
    dispatcher.add_listener("data.submit", validate_data)
    dispatcher.add_listener("data.submit", process_data)
    dispatcher.add_listener("data.submit", save_data)

    # First dispatch: validation fails, propagation stops
    print("=== Attempting to submit invalid data ===")
    from whistle import Event

    invalid_event = Event()
    invalid_event.valid = False
    dispatcher.dispatch("data.submit", invalid_event)

    print("\n=== Attempting to submit valid data ===")
    valid_event = Event()
    valid_event.valid = True
    dispatcher.dispatch("data.submit", valid_event)


if __name__ == "__main__":
    main()



"""
Custom event classes.

This example demonstrates creating custom Event subclasses to carry
domain-specific data and behavior.
"""

from whistle import Event, EventDispatcher


class UserRegisteredEvent(Event):
    """Custom event for user registration with domain data."""

    def __init__(self, user_id, username, email):
        """Initialize event with user data."""
        self.user_id = user_id
        self.username = username
        self.email = email

    def get_display_name(self):
        """Custom method on the event."""
        return f"{self.username} ({self.email})"


def main():
    """Demonstrate custom event classes."""
    dispatcher = EventDispatcher()

    # Listener that accesses custom event data
    def send_welcome_email(event):
        print(f"Sending welcome email to {event.email}")
        print(f"  User: {event.get_display_name()}")
        print(f"  ID: {event.user_id}")

    def log_registration(event):
        print(f"New user registered: {event.username}")

    # Register listeners
    dispatcher.add_listener("user.registered", send_welcome_email)
    dispatcher.add_listener("user.registered", log_registration)

    # Create and dispatch custom event
    event = UserRegisteredEvent(user_id=12345, username="alice", email="alice@example.com")
    dispatcher.dispatch("user.registered", event)


if __name__ == "__main__":
    main()



from unittest import mock

import pytest

from whistle import Event, EventDispatcher

STRING_EVENT_ID = "42"
OBJECT_EVENT_ID = object()
NUMERIC_EVENT_ID = 42


# todo deprecated: although it still works, event ids should be strings
BUNCH_OF_EVENT_IDS = [STRING_EVENT_ID, OBJECT_EVENT_ID, NUMERIC_EVENT_ID]


class CustomEvent(Event):
    pass


def parametrize_with_event_id():
    return pytest.mark.parametrize("event_id", BUNCH_OF_EVENT_IDS)


def test_event():
    event = Event()
    assert not event.propagation_stopped

    event.stop_propagation()
    assert event.propagation_stopped


@parametrize_with_event_id()
def test_dispatcher(event_id):
    dispatcher = EventDispatcher()

    listener = mock.MagicMock()
    assert not dispatcher.has_listeners(event_id)
    dispatcher.add_listener(event_id, listener)
    assert dispatcher.has_listeners(event_id)

    assert listener.call_count == 0

    e = dispatcher.dispatch(event_id)

    assert listener.call_count == 1
    assert not e.propagation_stopped

    e = dispatcher.dispatch(event_id)

    assert listener.call_count == 2
    assert not e.propagation_stopped


@pytest.mark.parametrize("EventType", [Event, CustomEvent])
@parametrize_with_event_id()
def test_dispatcher_custom_event(EventType, event_id):
    dispatcher = EventDispatcher()
    listener = mock.MagicMock()
    event = EventType()

    dispatcher.add_listener(event_id, listener)
    e = dispatcher.dispatch(event_id, event)

    assert listener.call_count == 1
    assert e == event


def test_propagation():
    dispatcher = EventDispatcher()

    for event_id in BUNCH_OF_EVENT_IDS:
        listener1 = mock.MagicMock()

        def listener2(event):
            event.stop_propagation()

        listener3 = mock.MagicMock()

        dispatcher.add_listener(event_id, listener1)
        dispatcher.add_listener(event_id, listener2)
        dispatcher.add_listener(event_id, listener3)

        assert listener1.call_count == 0
        assert listener3.call_count == 0

        e = dispatcher.dispatch(event_id)

        assert listener1.call_count == 1
        assert listener3.call_count == 0
        assert e.propagation_stopped

        e = dispatcher.dispatch(event_id)

        assert listener1.call_count == 2
        assert listener3.call_count == 0
        assert e.propagation_stopped

        assert dispatcher.get_listeners(event_id) == (listener1, listener2, listener3)

    listeners = dispatcher.get_listeners()
    assert len(listeners) == 3
    for event_id in BUNCH_OF_EVENT_IDS:
        assert len(listeners[event_id]) == 3


def test_no_listener():
    dispatcher = EventDispatcher()
    for event_id in BUNCH_OF_EVENT_IDS:
        assert not dispatcher.has_listeners(event_id)

        e = dispatcher.dispatch(event_id)
        assert not e.propagation_stopped


def test_remove_listener():
    dispatcher = EventDispatcher()

    for event_id in BUNCH_OF_EVENT_IDS:
        listener = mock.MagicMock()
        assert not dispatcher.has_listeners(event_id)
        dispatcher.add_listener(event_id, listener)
        assert dispatcher.has_listeners(event_id)

        assert listener.call_count == 0

        e = dispatcher.dispatch(event_id)

        assert listener.call_count == 1
        assert not e.propagation_stopped

        assert dispatcher.has_listeners(event_id)
        dispatcher.remove_listener(event_id, listener)
        assert not dispatcher.has_listeners(event_id)

        e = dispatcher.dispatch(event_id)

        assert listener.call_count == 1
        assert not e.propagation_stopped


@parametrize_with_event_id()
def test_listen_decorator(event_id):
    dispatcher = EventDispatcher()
    listener = mock.MagicMock()

    dispatcher.listen(event_id)(listener)
    e = dispatcher.dispatch(event_id)

    assert listener.call_count == 1
    assert not e.propagation_stopped


@parametrize_with_event_id()
def test_first_event_will_get_reference_to_dispatcher_and_name(event_id):
    dispatcher = EventDispatcher()
    listener = mock.MagicMock()

    dispatcher.add_listener(event_id, listener)
    e = dispatcher.dispatch(event_id)

    assert listener.call_count == 1
    assert not e.propagation_stopped
    assert e.dispatcher == dispatcher
    assert e.name == event_id



"""
Listener priority system.

This example demonstrates how to control the execution order of listeners
using priorities. Lower priority numbers execute first (higher priority).

Priority range: -127 to 128
Convention: -20 (highest) to 20 (lowest), default is 0
"""

from whistle import EventDispatcher


def main():
    """Demonstrate listener execution order with priorities."""
    dispatcher = EventDispatcher()

    # Register listeners with different priorities
    @dispatcher.listen("system.startup", priority=10)
    def low_priority_task(event):
        print("3. Low priority task (priority=10)")

    @dispatcher.listen("system.startup", priority=0)
    def normal_priority_task(event):
        print("2. Normal priority task (priority=0, default)")

    @dispatcher.listen("system.startup", priority=-10)
    def high_priority_task(event):
        print("1. High priority task (priority=-10)")

    # Dispatch the event - listeners execute in priority order
    print("Starting system...")
    dispatcher.dispatch("system.startup")
    print("System started")


if __name__ == "__main__":
    main()



"""
Asynchronous event dispatching with I/O operations.

This example shows how AsyncEventDispatcher is useful for coordinating
async operations like database writes, API calls, or file I/O.
"""

import asyncio

from whistle import AsyncEventDispatcher


async def main():
    """Demonstrate async event dispatching with simulated I/O operations."""
    dispatcher = AsyncEventDispatcher()

    # Simulate async database write
    async def save_to_database(event):
        print("Saving to database...")
        await asyncio.sleep(0.02)  # Simulate I/O delay
        print("Database save complete")

    # Simulate async API call
    async def notify_external_service(event):
        print("Notifying external service...")
        await asyncio.sleep(0.01)  # Simulate network delay
        print("External service notified")

    # Register async listeners
    dispatcher.add_listener("order.created", save_to_database)
    dispatcher.add_listener("order.created", notify_external_service)

    # Dispatch event - listeners execute sequentially with async/await
    print("Order created, processing...")
    await dispatcher.adispatch("order.created")
    print("All async operations complete")


if __name__ == "__main__":
    asyncio.run(main())



"""
Multiple listeners for a single event.

This example shows how multiple listeners can respond to the same event,
executing in the order they were registered (when no priority is specified).
"""

from whistle import EventDispatcher


def main():
    """Demonstrate multiple listeners for a single event."""
    dispatcher = EventDispatcher()

    # Define multiple listener functions
    def turn_off_lights(event):
        print("1. Turning off the lights")

    def start_music(event):
        print("2. Starting the music")

    def open_curtains(event):
        print("3. Opening the curtains")

    # Register all listeners for the same event
    dispatcher.add_listener("show.start", turn_off_lights)
    dispatcher.add_listener("show.start", start_music)
    dispatcher.add_listener("show.start", open_curtains)

    # Dispatch the event - all listeners will be called
    dispatcher.dispatch("show.start")


if __name__ == "__main__":
    main()
